/*-
 * Copyright (c) 2017 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* for strdup() */
#define _POSIX_C_SOURCE 200809L

#include "config.h"

#include <stdlib.h>
#include <string.h>

#include "cache/cache.h"
#include "vcl.h"
#include "vbm.h"

#include "vcc_if.h"

#define ERR(ctx, msg)					\
	VRT_fail((ctx), "vmod dispatch error: " msg)

#define VERR(ctx, fmt, ...)				\
	VRT_fail((ctx), "vmod dispatch error: " fmt, __VA_ARGS__)

#define INIT_BITS (64)

struct vmod_dispatch_label {
	unsigned	magic;
#define VMOD_DISPATCH_LABEL_MAGIC 0x44515cd0
	VCL_VCL		*vcl;
	struct vbitmap	*bitmap;
	char		*vcl_name;
	unsigned	nvcls;
};

static struct vrt_ctx dummy_ctx = { .magic = VRT_CTX_MAGIC };

VCL_VOID
vmod_label__init(VRT_CTX, struct vmod_dispatch_label **labelp,
		 const char *vcl_name)
{
	struct vmod_dispatch_label *label;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	AN(labelp);
	AZ(*labelp);
	AN(vcl_name);

	ALLOC_OBJ(label, VMOD_DISPATCH_LABEL_MAGIC);
	AN(label);
	*labelp = label;
	label->bitmap = vbit_new(INIT_BITS);
	AN(label->bitmap);
	label->vcl_name = strdup(vcl_name);
	AN(label->vcl_name);
	AZ(label->vcl);
	AZ(label->nvcls);
}

VCL_VOID
vmod_label__fini(struct vmod_dispatch_label **labelp)
{
	struct vmod_dispatch_label *label;

	if (labelp == NULL || *labelp == NULL)
		return;
	CHECK_OBJ(*labelp, VMOD_DISPATCH_LABEL_MAGIC);
	label = *labelp;
	*labelp = NULL;
	for (unsigned i = 0; i < label->nvcls; i++)
		if (vbit_test(label->bitmap, i) && label->vcl[i] != NULL)
			VRT_vcl_rel(&dummy_ctx, label->vcl[i]);
	if (label->bitmap != NULL)
		vbit_destroy(label->bitmap);
	if (label->vcl_name != NULL)
		free(label->vcl_name);
	FREE_OBJ(label);
}

VCL_VOID
vmod_label_add(VRT_CTX, struct vmod_dispatch_label *label, VCL_INT n,
	       VCL_STRING label_name)
{
	VCL_VCL vcl;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(label, VMOD_DISPATCH_LABEL_MAGIC);

	if ((ctx->method & VCL_MET_INIT) == 0) {
		VERR(ctx, "%s.add(%ld, %s) may only be called in vcl_init",
		     label->vcl_name, n, label_name);
		return;
	}
	if (n < 0) {
		VERR(ctx, "%s.add(%ld, %s): n must be >= 0", label->vcl_name,
		     n, label_name);
		return;
	}
	if ((vcl = VRT_vcl_get(ctx, label_name)) == NULL) {
		/*
		 * XXX: currently doesn't work, VRT_vcl_get() internally
		 * has AN(vcl);
		 */
		VERR(ctx, "%s.add(%ld, %s): unknown label", label->vcl_name, n,
		     label_name);
		return;
	}
	if (n >= label->nvcls)
		label->nvcls = n + 1;
	label->vcl = realloc(label->vcl, label->nvcls * sizeof(vcl));
	if (label->vcl == NULL) {
		VERR(ctx, "%s.add(%ld, %s): out of memory", label->vcl_name, n,
		     label_name);
		return;
	}
	label->vcl[n] = vcl;
	vbit_set(label->bitmap, n);
}

VCL_VOID
vmod_label_go(VRT_CTX, struct vmod_dispatch_label *label, VCL_INT n)
{
	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(label, VMOD_DISPATCH_LABEL_MAGIC);

	if (ctx->req == NULL) {
		VERR(ctx, "%s.go(%ld): may only be called in client context, "
		     "not in any vcl_backend_* subroutine", label->vcl_name,
		     n);
		return;
	}
	if (n < 0) {
		VERR(ctx, "%s.go(%ld): n must be >= 0", label->vcl_name, n);
		return;
	}
	if (n >= label->nvcls) {
		VERR(ctx, "%s.go(%ld): highest label number is %d",
		     label->vcl_name, n, label->nvcls - 1);
		return;
	}
	if (!vbit_test(label->bitmap, n)) {
		VERR(ctx, "%s.go(%ld): label %ld was not added",
		     label->vcl_name, n, n);
		return;
	}
	VRT_vcl_select(ctx, label->vcl[n]);
	VRT_handling(ctx, VCL_RET_VCL);
}

VCL_STRING
vmod_version(VRT_CTX)
{
	(void) ctx;
	return VERSION;
}
